import warnings
from abc import ABC, abstractmethod
from typing import List, Union, Tuple

from numpy import nextafter
from sklearn.base import BaseEstimator

from mindfoundry.optaas.client.expressions import Constraint, PriorMeanExpression
from mindfoundry.optaas.client.parameter import GroupParameter, Parameter

SMALLEST_NUMBER_ABOVE_ZERO = nextafter(0.0, 1)
LARGEST_NUMBER_BELOW_ONE = nextafter(1.0, 0)

ParametersConstraintsAndPriorMeans = Tuple[List[Parameter], List[Constraint], List[PriorMeanExpression]]


class Optimizable(ABC):
    """Superclass for all optimizable steps."""

    @abstractmethod
    def make_all_parameters_constraints_and_prior_means(self, estimator_name: str, id_prefix: str,
                                                        **kwargs) -> ParametersConstraintsAndPriorMeans:
        """Returns all parameters, constraints, and prior Means for multiple estimators in a group or choice."""


Estimator = Union[BaseEstimator, Optimizable]
EstimatorTuple = Tuple[str, Estimator]


def _get_all_parameters_and_constraints_and_prior_means(estimators: List[EstimatorTuple], id_prefix: str,
                                                        **kwargs) -> ParametersConstraintsAndPriorMeans:
    all_parameters = []
    all_constraints = []
    all_prior_means = []

    for estimator_name, estimator in estimators:
        if isinstance(estimator, Optimizable):
            parameters, constraints, prior_means\
                = estimator.make_all_parameters_constraints_and_prior_means(estimator_name, id_prefix, **kwargs)
        else:
            warnings.warn(f"{type(estimator).__name__} is not an Optimizable estimator.")
            parameters = [GroupParameter(estimator_name, id=_make_id(id_prefix + estimator_name), items=[])]
            constraints = []
            prior_means = []

        all_parameters.extend(parameters)
        all_constraints.extend(constraints)
        all_prior_means.extend(prior_means)

    return all_parameters, all_constraints, all_prior_means


def _make_id(estimator_name: str) -> str:
    return estimator_name.replace(' ', '-')


class MissingArgumentError(ValueError):
    """Raised when a required argument is missing from kwargs in :meth:`.OPTaaSClient.create_sklearn_task`"""

    def __init__(self, required_arg: str, estimator) -> None:
        super().__init__(f"{required_arg} is required in kwargs in order to optimize {type(estimator).__name__}")



import copy
from typing import Dict, Tuple, Any
import numpy as np
import pandas as pd
import holoviews as hv

from mindfoundry.optaas.client.client import OPTaaSClient
from mindfoundry.optaas.client.task import Task


class SurrogateViz:
    def __init__(self, client: OPTaaSClient, task: Task) -> None:
        self.client = client
        self.task = task
        self.continuous_vars: Dict[str, hv.Dimension] = {
            param["name"]: hv.Dimension(
                param["name"], range=(param["minimum"], param["maximum"])
            )
            for param in task.parameters
            if param["type"] == "number"
        }
        self.categorical_vars: Dict[str, hv.Dimension] = {
            param["name"]: hv.Dimension(param["name"], values=param["enum"])
            for param in task.parameters
            if param["type"] == "categorical"
        }

        self.grid_size: int = 30

    def plot_surrogate_mean_and_std(
            self,
            x_var: str,
            y_var: str,
            z_range: Tuple[float, float],
            fig_width: int = 600,
            fig_height: int = 600,
        ) -> hv.DynamicMap:
        if not x_var in self.continuous_vars:
            raise KeyError(f"Must choose one of {self.continuous_vars}.\n Got {x_var}")
        if not y_var in self.continuous_vars:
            raise KeyError(f"Must choose one of {self.continuous_vars}.\n Got {y_var}")

        k_dims = list(self.categorical_vars.values()) + [
            dim
            for dim in self.continuous_vars.values()
            if dim.name not in (x_var, y_var)
        ]

        def make_surface(*args, **kwargs) -> hv.Layout:
            # Dangerous, assume the args are in the same order as in k_dims
            kwargs = {dim.name: value for dim, value in zip(k_dims, args)}

            _, _, mean, variance = self._predict_at_slice(
                x_var=x_var, y_var=y_var, fixed_vars=kwargs
            )
            bounds = (
                self.continuous_vars[x_var].range[0],
                self.continuous_vars[y_var].range[0],
                self.continuous_vars[x_var].range[1],
                self.continuous_vars[y_var].range[1],
            )
            mean_surface = hv.Surface(mean, bounds=bounds).opts(
                title="Surrogate Mean", xlabel=x_var, ylabel=y_var, zlabel="Value"
            )
            std_surface = hv.Surface(np.sqrt(variance), bounds=bounds).opts(
                title="Surrogate Standard Deviation",
                xlabel=x_var,
                ylabel=y_var,
                zlabel="Value",
            )
            layout = mean_surface + std_surface
            layout.cols(1)
            return layout

        dmap = (
            hv.DynamicMap(make_surface, kdims=k_dims)
            .opts(width=fig_width, height=fig_height)
            .redim.range(z=z_range)
        )
        return dmap

    def plot_surrogate_with_uncertainties(
            self,
            x_var: str,
            y_var: str,
            z_range: Tuple[float, float],
            fig_width: int = 600,
            fig_height: int = 600,
        ) -> hv.DynamicMap:
        if not x_var in self.continuous_vars:
            raise KeyError(f"Must choose one of {self.continuous_vars}.\n Got {x_var}")
        if not y_var in self.continuous_vars:
            raise KeyError(f"Must choose one of {self.continuous_vars}.\n Got {y_var}")

        k_dims = list(self.categorical_vars.values()) + [
            dim
            for dim in self.continuous_vars.values()
            if dim.name not in (x_var, y_var)
        ]

        def make_surface(*args, **kwargs) -> hv.Layout:
            # Dangerous, assume the args are in the same order as in k_dims
            kwargs = {dim.name: value for dim, value in zip(k_dims, args)}

            _, _, mean, variance = self._predict_at_slice(
                x_var=x_var, y_var=y_var, fixed_vars=kwargs
            )
            bounds = (
                self.continuous_vars[x_var].range[0],
                self.continuous_vars[y_var].range[0],
                self.continuous_vars[x_var].range[1],
                self.continuous_vars[y_var].range[1],
            )

            mean_surface = hv.Surface(mean, bounds=bounds)
            upper_surface = hv.Surface(
                mean + 2 * np.sqrt(variance), bounds=bounds
            ).options(alpha=0.4, cmap=["blue", "blue"])
            lower_surface = hv.Surface(
                mean - 2 * np.sqrt(variance), bounds=bounds
            ).options(alpha=0.4, cmap=["blue", "blue"])
            layout = (
                mean_surface
                * upper_surface
                * lower_surface.opts(
                    title="Surrogate Mean +- 2 std",
                    xlabel=x_var,
                    ylabel=y_var,
                    zlabel="Value",
                )
            )
            return layout

        dmap = (
            hv.DynamicMap(make_surface, kdims=k_dims)
            .opts(width=fig_width, height=fig_height)
            .redim.range(z=z_range)
        )
        return dmap

    def _predict_at_slice(self, x_var: str, y_var: str, fixed_vars: Dict[str, Any]):
        var_dict = copy.copy(fixed_vars)
        x_grid, y_grid = self._make_grid(x_var=x_var, y_var=y_var)
        var_dict[x_var], var_dict[y_var] = x_grid.ravel(), y_grid.ravel()
        query_df = pd.DataFrame(data=var_dict)
        prediction = self._query_surrogate(data=query_df)

        return (
            x_grid,
            y_grid,
            prediction["mean"].values.reshape(x_grid.shape),
            prediction["variance"].values.reshape(y_grid.shape),
        )

    def _make_grid(self, x_var: str, y_var: str) -> Tuple[np.ndarray, np.ndarray]:
        x_range = np.linspace(
            self.continuous_vars[x_var].range[0],
            self.continuous_vars[x_var].range[1],
            self.grid_size,
        )
        y_range = np.linspace(
            self.continuous_vars[y_var].range[0],
            self.continuous_vars[y_var].range[1],
            self.grid_size,
        )
        x_grid, y_grid = np.meshgrid(x_range, y_range)
        return x_grid, y_grid

    def _query_surrogate(self, data: pd.DataFrame) -> pd.DataFrame:
        query = data.to_dict(orient="records")
        predictions = self.task.get_surrogate_predictions(query)
        raw_mean, raw_variance = zip(
            *[(pred.mean, pred.variance) for pred in predictions]
        )
        result = pd.DataFrame(
            index=data.index, data={"mean": raw_mean, "variance": raw_variance}
        )
        return result



from typing import Any, Dict

from mindfoundry.optaas.client.goal import Goal


class Objective:
    """An objective for optimizing as part of a multi-objective Task.

    When evaluating a :class:`.Configuration` and producing a :class:`Result`, you can include a score for each Objective.

    Args:
        id (str): A unique id for the objective.
        goal (Goal, optional, default Goal.max): Whether OPTaaS should aim for the lowest (min) or highest (max) score.
        min_known_score (float, optional): Minimum known score value for this objective.
        max_known_score (float, optional): Maximum known score value for this objective.
    """

    def __init__(self, id: str,  # pylint: disable=redefined-builtin
                 goal: Goal = None, min_known_score: float = None, max_known_score: float = None) -> None:
        self._id = id
        self._goal = goal
        self._min_known_score = min_known_score
        self._max_known_score = max_known_score

    def to_json(self) -> Dict[str, Any]:
        json: Dict[str, Any] = {"id": self._id}
        if self._goal:
            json["goal"] = self._goal.name
        if self._min_known_score is not None:
            json["minKnownScore"] = self._min_known_score
        if self._max_known_score is not None:
            json["maxKnownScore"] = self._max_known_score
        return json



from sklearn.ensemble import AdaBoostClassifier as BaseAdaBoostClassifier, AdaBoostRegressor as BaseAdaBoostRegressor

from mindfoundry.optaas.client.parameter import Distribution
from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersConstraintsAndPriorMeans


class _OptimizableAdaBoost(OptimizableBaseEstimator):
    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) \
            -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`,
        and :class:`PriorMeans <.PriorMeans>` to optimize an AdaBoost estimator."""

        return [
            sk.IntParameter('n_estimators', minimum=10, maximum=500),
            sk.FloatParameter('learning_rate', minimum=0.01, maximum=2.5, distribution=Distribution.LOGUNIFORM),
        ], [], []


class AdaBoostClassifier(BaseAdaBoostClassifier, _OptimizableAdaBoost):
    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) \
            -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`,
        and :class:`PriorMeans <.PriorMeans>` to optimize a :class:`.AdaBoostClassifier`."""

        parameters, constraints, prior_means = super().make_parameters_constraints_and_prior_means(sk, **kwargs)
        parameters.append(sk.CategoricalParameter('algorithm', values=['SAMME', 'SAMME.R']))
        return parameters, constraints, prior_means


class AdaBoostRegressor(BaseAdaBoostRegressor, _OptimizableAdaBoost):
    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) \
            -> ParametersConstraintsAndPriorMeans:

        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`,
        and :class:`PriorMeans <.PriorMeans>` to optimize a :class:`.AdaBoostRegressor`."""

        parameters, constraints, prior_means = super().make_parameters_constraints_and_prior_means(sk, **kwargs)
        parameters.append(sk.CategoricalParameter('loss', values=['linear', 'square', 'exponential']))
        return parameters, constraints, prior_means



from typing import Any, Dict, Union, Tuple

from mindfoundry.optaas.client.configuration import Configuration
from mindfoundry.optaas.client.utils import _pprint

ScoreValue = float
ScoreDict = Dict[str, float]
ScoreValueOrDict = Union[ScoreValue, ScoreDict]
VarianceValue = float
VarianceDict = Dict[str, float]
VarianceValueOrDict = Union[VarianceValue, VarianceDict]
ScoringFunctionOutput = Union[ScoreValueOrDict, Tuple[ScoreValue, VarianceValue], Tuple[ScoreDict, VarianceDict]]


class Result:
    """The result obtained by taking a :class:`.Configuration` and running it through your scoring function.

    Must contain **either** a `score` value **or** an `error`.
    You cannot specify `variance` without `score`.

    Args:
        configuration (Configuration): The Configuration used to obtain this Result.
        score (ScoreValueOrDict): The score obtained.
            For multi-objective tasks, a dictionary of scores for each objective (using the objective IDs as keys).
        variance (VarianceValueOrDict >=0, optional, defaults to 0): Variance associated with the score.
            For multi-objective tasks, a dictionary of variances for each objective (using the objective IDs as keys).
        error (Any): Any data related to an error encountered while calculating the score.
        user_defined_data (Any, optional): Any other data you wish to store as part of this Result.

    Raises:
        :class:`.ValueError` if invalid data is provided (e.g. both score and error are provided, or neither).
    """

    def __init__(self, configuration: Configuration, score: Union[float, Dict] = None, error: str = None,
                 variance: Union[float, Dict] = None, user_defined_data: Any = None) -> None:
        if (score is None) == (error is None):
            raise ValueError("Results must specify either a score or an error, not both.")

        if score is None and variance is not None:
            raise ValueError("Cannot specify variance without score.")

        self.configuration = configuration
        self.score = score
        self.error = error
        self.variance = variance
        self.user_defined_data = user_defined_data

    def to_json(self) -> Dict[str, Any]:
        json = self.to_json_without_configuration()
        json['configuration'] = self.configuration.id
        return json

    def to_json_without_configuration(self) -> Dict[str, Any]:
        json: Dict[str, Any] = {'score': self.score} if self.error is None else {'error': self.error}
        if self.variance is not None:
            json['variance'] = self.variance
        if self.user_defined_data is not None:
            json['userDefined'] = self.user_defined_data
        return json


class StoredResult:
    """A Result that has been stored in OPTaaS.

    Attributes:
        json (Dict): The full Result object as stored in OPTaaS in JSON format.
        id (int): Unique id for the Result.
        configuration (Configuration):
            The :class:`.Configuration` used to obtain the Result
        pipeline (Optional[Pipeline]):
            The Pipeline corresponding to the Configuration (only available for SklearnTasks).
        score (ScoreValueOrDict): Score obtained by using the :class:`.Configuration`, or a Dict of scores for each objective.
        variance (VarianceValueOrDict): Variance associated with the score, or a Dict of variance for each objective.
        error (Any): Any data related to an error encountered when trying to use the :class:`.Configuration`.
        user_defined_data (Any): Any other data provided when storing the Result.
    """

    def __init__(self, json: Dict) -> None:
        self.json = json
        self.id = json['id']  # pylint: disable=invalid-name
        self.configuration = Configuration(json['configuration'])
        self.pipeline = None
        self.score = json.get('score')
        self.variance = json.get('variance')
        self.error = json.get('error')
        self.user_defined_data = json.get('userDefined')

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __repr__(self):
        if self.pipeline is None:
            return _pprint(self, 'configuration', 'score', 'user_defined_data')
        return _pprint(self, 'pipeline', 'score', 'user_defined_data')

    def as_pandas_row(self) -> Dict:
        """Returns this Result as a Dict that can be used to create a Pandas DataFrame."""
        return {
            'config': self.configuration.values,
            'score': self.score,
            'variance': self.variance,
            'error': self.error
        }



from typing import List, Any, Dict, TYPE_CHECKING

from mindfoundry.optaas.client.api_key import ApiKey, UserRole
from mindfoundry.optaas.client.expressions import PriorMeanExpression, Constraint
from mindfoundry.optaas.client.goal import Goal
from mindfoundry.optaas.client.utils import MfOptOptimizer
from mindfoundry.optaas.client.objective import Objective
from mindfoundry.optaas.client.parameter import Parameter, GroupParameter
from mindfoundry.optaas.client.session import OPTaaSSession, DEFAULT_MAX_RETRIES
from mindfoundry.optaas.client.task import Task
from sympy import Expr

if TYPE_CHECKING:  # pragma: no cover
    from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizablePipeline  # pylint: disable=unused-import
    from mindfoundry.optaas.client.sklearn_pipelines.sklearn_task import SklearnTask  # pylint: disable=unused-import

_API_ROOT = '/api/v1'


class OPTaaSClient:
    """Sets up a connection to OPTaaS and allows you to create a :class:`.Task`, retrieve existing Tasks etc.

    Args:
        server_url (str): URL of your OPTaaS server
        api_key (str): Your personal API key
        disable_version_check (bool, default False):
            Set to True if you don't want to be notified when a new version of the client is available
        max_retries (int, default 3): How many times to retry a request if a connection error occurs.
        keep_alive (bool, default True): Very rarely required. Set to False only if you experience connection dropping issues.
    """

    def __init__(self, server_url: str, api_key: str,
                 disable_version_check: bool = False, max_retries: int = DEFAULT_MAX_RETRIES, keep_alive: bool = True) -> None:
        self._session = OPTaaSSession(server_url=server_url, api_key=api_key, disable_version_check=disable_version_check,
                                      max_retries=max_retries, keep_alive=keep_alive)
        root_response = self._session.get(_API_ROOT, timeout=2)
        self._tasks_endpoint = root_response.body['_links']['tasks']['href']
        self._api_key_endpoint = root_response.body['_links']['apiKeys']['href']

    def _create_task(self, title: str, parameters: List[Parameter], constraints: List[Constraint] = None,
                    prior_means: List[PriorMeanExpression] = None, random_seed: int = None,
                    initial_configurations: int = None, objectives: List[Objective] = None, goal: Goal = None,
                    min_known_score: float = None, max_known_score: float = None, user_defined_data: Any = None,
                    targets: List[Parameter] = None, likelihood: Expr = None, optimizer_name: str = None)\
            -> Task:
        """Formats the body of the request and returns a new :class: `.Task` by making a POST request to OPTaaS"""
        body: Dict = dict(
            title=title,
            parameters=[p.to_json() for p in parameters],
            constraints=[constraint.to_optaas_expression() for constraint in constraints] if constraints else [],
            priorMeans=[prior_mean.to_optaas_expression() for prior_mean in prior_means] if prior_means else [],
            objectives=[objective.to_json() for objective in objectives] if objectives else None,
            goal=goal.name if goal else None,
            randomSeed=random_seed,
            userDefined=user_defined_data,
            initialConfigurations=initial_configurations,
            minKnownScore=min_known_score,
            maxKnownScore=max_known_score,
            targets=[t.to_json() for t in targets] if targets else None,
            likelihood=str(likelihood) if likelihood else None,
            optimizerName=optimizer_name
        )
        body = {key: value for key, value in body.items() if value is not None}
        response = self._session.post(self._tasks_endpoint, body=body)
        return Task(json=response.body, session=self._session)

    def create_task(self, title: str, parameters: List[Parameter], constraints: List[Constraint] = None,
                    prior_means: List[PriorMeanExpression] = None, random_seed: int = None,
                    initial_configurations: int = None, objectives: List[Objective] = None, goal: Goal = None,
                    min_known_score: float = None, max_known_score: float = None, user_defined_data: Any = None)\
            -> Task:
        """Creates a new :class:`.Task` by making a POST request to OPTaaS

        Args:
            title (str): Name/description of your Task.
            parameters (List[Parameter]): Parameters that you would like to optimize.
            constraints (List[Constraint]): Constraints on what values can be assigned to Parameters.
            prior_means (List[PriorMeansExpression]): Prior means to be used by the optimizer.
                This is a list of expressions provided by the user which together define a guess for the function we are
                trying to optimize. Expressions are provided in the same "When-Then" format that is used to describe input
                parameter constraints. The ordering of this list is important: the first expression whose "When" evaluates
                to True will be used to calculate the prior mean value. If none evaluate to True, a default value of 0 will
                be used, which is equivalent to there being no prior mean known. Prior means do not come with uncertainty
                information, and they are directly subtracted from the data before a surrogate model is fitted.
            initial_configurations (int, optional, default 10, minimum 1):
                Number of Configurations that OPTaaS will generate upfront. If you are planning to have multiple clients
                working concurrently, set this to be equal to the number of clients.
            objectives (List[Objective], optional): Specification of each objective for a multi-objective optimization.
                Note: Conditional parameters (i.e. ChoiceParameters and anything with optional=True) are not supported
                in multi-objective tasks.
            goal (Goal, optional, default Goal.max): Whether OPTaaS should aim for the lowest (min) or highest (max) score.
                Do not use in multi-objective optimizations (specify goal for each objective instead).
            min_known_score (float, optional): Minimum known score value.
                Do not use in multi-objective optimizations (specify min_known_score for each objective instead).
            max_known_score (float, optional): Maximum known score value.
                Do not use in multi-objective optimizations (specify max_known_score for each objective instead).
            user_defined_data (Any, optional): Any other data you would like to store in the task JSON
            random_seed (int, optional):
                Seed for the random generator used by OPTaaS when generating :class:`Configurations <.Configuration>`.
                If not specified, a new seed will be used for each Task.
                Use this only if you need reproducible results, i.e. if you create 2 Tasks with identical attributes
                including an identical `random_seed`, and you use the same scoring function, then OPTaaS is guaranteed
                to generate the same Configurations in the same order for both Tasks.

        Returns:
            A new :class:`.Task`

        Raises:
            :class:`.OPTaaSError` if the Task data is invalid or the server is unavailable.
        """
        return self._create_task(
            title=title,
            parameters=parameters,
            constraints=constraints,
            prior_means=prior_means,
            objectives=objectives,
            goal=goal,
            random_seed=random_seed,
            user_defined_data=user_defined_data,
            initial_configurations=initial_configurations,
            min_known_score=min_known_score,
            max_known_score=max_known_score,
            optimizer_name=MfOptOptimizer.BAYESIAN.value
        )

    def create_likelihood_task(self, title: str, parameters: List[Parameter],
                    targets: List[Parameter],
                    observed_data: List[Objective],
                    likelihood: Expr,
                    constraints: List[Constraint] = None,
                    random_seed: int = None,
                    initial_configurations: int = None,
                    user_defined_data: Any = None,)\
            -> Task:
        """Creates a new :class:`.Task` by making a POST request to OPTaaS

        Args:
            title (str): Name/description of your Task.
            parameters (List[Parameter]): Parameters that you would like to optimize.
            constraints (List[Constraint]): Constraints on what values can be assigned to Parameters.
            targets (List[Parameter]): Parameters to be optimized by maximising the likelihood.
                Note 1: Currently only one single is supported.
            observed_data (List[Objective], optional): Specification of the observed data at the configurations.
                Note: Conditional parameters (i.e. ChoiceParameters and anything with optional=True) are not supported
                in multi-objective tasks.
            likelihood (Expr): The likelihood function expressed as sympy Expr to be optimized wrt to the targets' variables.
                The likelihood is in function of the parameters, observed_data and targets.
            initial_configurations (int, optional, default 10, minimum 1):
                Number of Configurations that OPTaaS will generate upfront. If you are planning to have multiple clients
                working concurrently, set this to be equal to the number of clients.
            user_defined_data (Any, optional): Any other data you would like to store in the task JSON
            random_seed (int, optional):
                Seed for the random generator used by OPTaaS when generating :class:`Configurations <.Configuration>`.
                If not specified, a new seed will be used for each Task.
                Use this only if you need reproducible results, i.e. if you create 2 Tasks with identical attributes
                including an identical `random_seed`, and you use the same scoring function, then OPTaaS is guaranteed
                to generate the same Configurations in the same order for both Tasks.

        Returns:
            A new :class:`.Task`

        Raises:
            :class:`.OPTaaSError` if the Task data is invalid or the server is unavailable.
        """
        return self._create_task(
            title=title,
            parameters=parameters,
            constraints=constraints,
            targets=targets,
            objectives=observed_data,
            likelihood=likelihood,
            random_seed=random_seed,
            user_defined_data=user_defined_data,
            initial_configurations=initial_configurations,
            optimizer_name=MfOptOptimizer.LIKELIHOOD.value,
        )

    def create_sklearn_task(self, title: str, pipeline: 'OptimizablePipeline',
                            additional_parameters: List[Parameter] = None,
                            additional_constraints: List[Constraint] = None,
                            additional_prior_means: List[PriorMeanExpression] = None,
                            random_seed: int = None, initial_configurations: int = None, objectives: List[Objective] = None,
                            goal: Goal = None, min_known_score: float = None, max_known_score: float = None,
                            user_defined_data: Any = None, **kwargs) -> 'SklearnTask':
        """Creates a new :class:`.SklearnTask` by making a POST request to OPTaaS

        All the arguments from :meth:`.OPTaaSClient.create_task` can be used here except instead of
        `parameters` and `constraints` there is `additional_parameters` and `additional_constraints`.

        Args:
            pipeline (OptimizablePipeline): The pipeline you wish to optimize.
            additional_parameters (List[Parameter], optional):
                Additional parameters that you would like to optimize.
            additional_constraints (List[Constraint], optional):
                Additional constraints on your Parameters.
            additional_prior_means (List[Constraint], optional):
                Additional prior_means on your Parameters.
            kwargs:
                Additional arguments required to optimize certain estimators, e.g. :class:`.PCA` requires `feature_count`.

        Returns:
            A new :class:`.SklearnTask`

        Raises:
            :class:`.MissingArgumentError` if a required argument is missing from `kwargs`.
            :class:`.OPTaaSError` if the Task data is invalid or the server is unavailable.
        """
        from mindfoundry.optaas.client.sklearn_pipelines.sklearn_task import SklearnTask  # pylint: disable=redefined-outer-name,import-outside-toplevel

        parameters, constraints, prior_means = pipeline.make_all_parameters_constraints_and_prior_means('pipeline', '', **kwargs)

        if additional_parameters:
            parameters.append(GroupParameter('additional', items=additional_parameters))
        if additional_constraints:
            constraints.extend(additional_constraints)
        if additional_prior_means:
            prior_means.extend(additional_prior_means)
        task = self._create_task(
            title=title,
            parameters=parameters,
            constraints=constraints,
            prior_means=prior_means,
            objectives=objectives,
            goal=goal,
            random_seed=random_seed,
            user_defined_data=user_defined_data,
            initial_configurations=initial_configurations,
            min_known_score=min_known_score,
            max_known_score=max_known_score,
        )

        return SklearnTask(task, pipeline.estimators)

    def get_all_tasks(self) -> List[Task]:
        """Retrieves a list of all stored Tasks by making a GET request to OPTaaS.

        Returns:
            List of :class:`Tasks <.Task>`

        Raises:
            :class:`.OPTaaSError` if the server is unavailable
        """
        response = self._session.get(self._tasks_endpoint)
        return [Task(json, self._session) for json in response.body['tasks']]

    def get_task(self, task_id: str) -> Task:
        """Retrieves a stored :class:`.Task` by making a GET request to OPTaaS.

        Args:
            task_id (str): unique id for the Task

        Returns:
            A :class:`.Task`

        Raises:
            :class:`.OPTaaSError` if no record is found with the given id or the server is unavailable.
        """
        response = self._session.get(f'{self._tasks_endpoint}/{task_id}')
        return Task(response.body, self._session)

    def get_sklearn_task(self, task_id: str, pipeline: 'OptimizablePipeline') -> 'SklearnTask':
        """Retrieves a stored :class:`.SklearnTask` by making a GET request to OPTaaS.

        This allows you to create a SklearnTask and then use it again in a separate/later session, assuming of course
        that you call this method with the same `estimators` you used to create the original task.

        Args:
            task_id (str): unique id for the Task
            pipeline (OptimizablePipeline): The same pipeline used when calling :meth:`.OPTaaSClient.create_sklearn_task`

        Returns:
            A :class:`.SklearnTask`

        Raises:
            :class:`.OPTaaSError` if no record is found with the given id or the server is unavailable.
        """
        from mindfoundry.optaas.client.sklearn_pipelines.sklearn_task import SklearnTask  # pylint: disable=redefined-outer-name,import-outside-toplevel
        task = self.get_task(task_id)
        return SklearnTask(task, pipeline.estimators)

    def get_api_keys(self) -> List[ApiKey]:
        """Retrieves the list of all API keys by making a GET request to OPTaaS. Only available to Admin users."""
        response = self._session.get(self._api_key_endpoint)
        return [ApiKey(json, self._session) for json in response.body['apiKeys']]

    def generate_api_key(self, role: UserRole = None) -> ApiKey:
        """Generates a new API key by making a POST request to OPTaaS. Only available to Admin users."""
        response = self._session.post(self._api_key_endpoint, {} if role is None else {"role": role.value})
        return ApiKey(response.body, self._session)



import sys
from typing import Any, List, Literal, Dict, Callable, Optional, Tuple, TypeVar, Union, TYPE_CHECKING

from loguru import logger
from mindfoundry.optaas.client.configuration import Configuration
from mindfoundry.optaas.client.user_defined_configuration import UserDefinedConfiguration
from mindfoundry.optaas.client.goal import Goal
from mindfoundry.optaas.client.prediction import Prediction
from mindfoundry.optaas.client.result import StoredResult, Result, ScoreValue, ScoreDict, ScoreValueOrDict, VarianceValueOrDict, \
    ScoringFunctionOutput
from mindfoundry.optaas.client.session import OPTaaSSession, OPTaaSResponse
from mindfoundry.optaas.client.utils import _pprint

if TYPE_CHECKING:  # pragma: no cover
    from pandas import DataFrame  # pylint: disable=unused-import

_MULTI_OBJECTIVE_UNSUPPORTED_ERROR = "best_first is not supported for multi-objective Tasks.\n" \
                                     "Did you mean get_pareto_set?"

_SINGLE_OBJECTIVE_UNSUPPORTED_ERROR = "get_pareto_set is not supported for single-objective Tasks.\n" \
                                      "Did you mean get_best_result_and_configuration?"

_NO_SCORE_THRESHOLD_MSG = "no score threshold set"

_UNTIL_SCORE_THRESHOLD_MSG = (
    lambda score_threshold: f"or until score is {score_threshold} or better"
)

_UNTIL_SCORES_THRESHOLD_MSG = (
    lambda score_threshold: f"or until scores are {score_threshold} or better"
)

T = TypeVar('T')

MaybeList = Union[T, List[T]]
LoggingLevel = Literal["DEBUG", "INFO", "WARNING", "ERROR"]

class Task:
    """Allows you to access Task attributes and perform all Task-related functions.

    Attributes:
        json (Dict): The full JSON representation of this Task in OPTaaS.
        id (str): Unique id for the Task.
        title (str): Name/description as provided when the Task was created.
        parameters (List[Dict]): JSON representation of the :class:`Parameters <.Parameter>` defined for this Task.
        constraints (List[str]): List of OPTaaS string representations of the :class:`Constraints <.Constraint>` defined for this Task.
        prior_means (List[str]): List of OPTaaS string representations of the :class:`Prior Means <.PriorMeanExpression>` defined for this Task.
        status (str): Current status of the Task, e.g. 'running' or 'done'
        number_of_iterations (int): Number of :class:`Results <.Result>` provided for this Task so far.
        user_defined_data (Any): Any other data as provided when the Task was created.
    """

    def __init__(self, json: dict, session: OPTaaSSession) -> None:
        self.json = json
        self.id = json['id']  # pylint: disable=invalid-name
        self.parameters = json['parameters']
        self.constraints = json['constraints']
        self.prior_means = json['priorMeans']
        self._update_attributes(json)

        self._session = session
        self._task_url = json['_links']['self']['href']
        self._configurations_url = json['_links']['configurations']['href']
        self._results_url = json['_links']['results']['href']
        self._complete_url = json['_links']['complete']['href']
        self._resume_url = json['_links']['resume']['href']
        self._predictions_url = json['_links']['predictions']['href']
        self._pareto_url = json['_links']['pareto']['href']
        self._targets_estimates_url = json['_links']['targets_estimates']['href']
        self._sqrt_crlb_url = json['_links']['sqrt_crlb']['href']
        self._handler_id: Optional[int] = None

    def _update_attributes(self, json: dict):
        self.title = json['title']
        self.status = json['status']
        self.number_of_iterations = json['numberOfIterations']
        self.user_defined_data = json.get('userDefined')

    def _setup_logging_level(self, logging_level: LoggingLevel) -> None:
        # remove default handler
        try:
            logger.remove(0)
        except ValueError:
            pass
        if self._handler_id is not None:
            logger.remove(self._handler_id)
        self._handler_id = logger.add(
            sys.stderr,
            filter=__name__.split(".", 1)[0],
            format="<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | <level>{level: <8}</level> | <cyan> Task.{function}</cyan> | <level>{message}</level>",
            level=logging_level,
            colorize=True,
        )

    def run(self, scoring_function: Callable[..., ScoringFunctionOutput], max_iterations: int,
            score_threshold: ScoreValueOrDict = None, logging_level: LoggingLevel = "INFO") -> Union[StoredResult, List[StoredResult]]:
        """Run this task, using the provided scoring function to calculate the score for each configuration.

        Args:
            scoring_function (Callable[..., ScoringFunctionOutput]):
                Function that takes configuration values as arguments, e.g. if you have parameters x and y, your
                function would be `def get_score(x, y)`.
                The function can return just a score, or a tuple of (score, variance).
                For multi-objective tasks the score and variance must be dictionaries, with Objective.id values as keys, e.g.
                `{"id1": 1.23, "id2": 4.56}`.
            max_iterations (int):
                Max number of iterations to run, i.e. number of results to record before stopping.
            score_threshold (ScoreValueOrDict, optional, defaults to min/max known score if defined):
                Stop running the task when the score is equal to or better than this value.
                For multi-objective tasks, use a dictionary with Objective.id values as keys, e.g.
                `{"id1": 1.23, "id2": 4.56}`.
            logging_level (Literal["DEBUG", "INFO", "WARNING", "ERROR"], optional, defaults to "INFO"): Set the logging level.

        Returns:
            The best recorded :class:`.Result` with the :class:`.Configuration` that was used to achieve it.
            For multi-objective tasks, the set of Pareto front Results will be returned instead.

        Raises:
            :class:`.OPTaaSError` if the server is unavailable.
        """
        self._setup_logging_level(logging_level)

        starting_msg = f'Running task "{self.title}" for {max_iterations} iterations'

        reached_threshold, msg = self._reached_threshold_if_defined(score_threshold)
        logger.info(f"{starting_msg} ({msg})\n")

        configuration = self.generate_configurations(1)[0]

        for i in range(max_iterations):
            scoring_result = scoring_function(**configuration.values)

            if isinstance(scoring_result, tuple):
                score, variance = scoring_result
                score_msg = (
                    f"Iteration: {i} | Score: {score} | Variance: {variance}"
                )
            else:
                score = scoring_result
                variance = None  # type: ignore
                score_msg = f"Iteration: {i} | Score: {score}"

            logger.info(f"{score_msg} | Configuration: {configuration.values}")

            configuration = self.record_result(configuration=configuration, score=score, variance=variance)
            if reached_threshold(score):  # type: ignore
                break

        return_value = self.get_pareto_set() if self.json.get('objectives') else self.get_best_result_and_configuration()

        self.complete()
        self.refresh()
        logger.info("Task Completed")

        return return_value  # type: ignore

    def _reached_threshold_if_defined(self, score_threshold: ScoreValueOrDict = None) -> Tuple[Union[Callable[[ScoreValue], bool],
                                                                                               Callable[[ScoreDict], bool]], str]:
        objectives = self.json.get('objectives')
        if objectives:
            if score_threshold is None or isinstance(score_threshold, Dict):
                return self._reached_multi_objective_threshold_if_defined(objectives, score_threshold or {})
            raise ValueError("Score threshold must be a dictionary")

        goal = Goal[self.json['goal']]

        if score_threshold is None:
            score_threshold = self.json.get('minKnownScore' if goal == Goal.min else 'maxKnownScore')
            if score_threshold is None:
                print('(no score threshold set)')
                return lambda score: False, _NO_SCORE_THRESHOLD_MSG

        print(f'(or until score is {score_threshold} or better)')
        if goal == Goal.min:
            return lambda score: score <= score_threshold, _UNTIL_SCORE_THRESHOLD_MSG(
                score_threshold
            )
        return lambda score: score >= score_threshold, _UNTIL_SCORE_THRESHOLD_MSG(
                score_threshold
            )

    @staticmethod
    def _reached_multi_objective_threshold_if_defined(objectives: List[Dict],
                                                      score_threshold: ScoreDict) -> Tuple[Callable[[ScoreDict], bool], str]:
        goals = {objective['id']: Goal[objective['goal']] for objective in objectives}

        for objective in objectives:
            id_ = objective['id']
            if id_ not in score_threshold:
                best_known_score = objective.get('minKnownScore' if goals[id_] == Goal.min else 'maxKnownScore')
                if best_known_score is not None:
                    score_threshold[id_] = best_known_score

        if score_threshold == {}:
            return lambda scores: False, _NO_SCORE_THRESHOLD_MSG

        print(f'(or until scores are {score_threshold} or better)')

        def reached_all_thresholds(scores: ScoreDict) -> bool:
            for id_, threshold in score_threshold.items():
                score = scores.get(id_)
                if score is None:
                    return False
                if goals[id_] == Goal.min:
                    if score > threshold:
                        return False
                elif score < threshold:
                    return False
            return True

        return reached_all_thresholds, _UNTIL_SCORES_THRESHOLD_MSG(score_threshold)

    def refresh(self):
        """Make a GET request to OPTaaS to retrieve the latest Task data and update this object accordingly."""
        response = self._session.get(self._task_url)
        self._update_attributes(response.body)

    def delete(self):
        """Delete this Task (cannot be undone)."""
        self._session.delete(self._task_url)

    def generate_configurations(self, quantity: int = 1) -> List[Configuration]:
        """Make a POST request to OPTaaS to generate a set of new :class:`Configurations <.Configuration>` for this Task.

        Args:
            quantity (int, optional, default 1): The number of configurations to generate (minimum 1).

        Returns:
            A list of the newly created :class:`Configurations <.Configuration>`.

        Raises:
            :class:`.OPTaaSError` if the server is unavailable or the quantity is invalid.
        """
        response = self._session.post(self._configurations_url + f'?quantity={quantity}', {})
        return [Configuration(json) for json in response.body['configurations']]

    def add_user_defined_configuration(
        self,
        values: Optional[Dict] = None,
        score: Optional[ScoreValueOrDict] = None,
        variance: Optional[VarianceValueOrDict] = None,
        user_defined_data: Any = None,
        configurations: Optional[List[UserDefinedConfiguration]] = None
        ) -> MaybeList[Configuration]:
        """Make a POST request to OPTaaS to store a list of user-provided :class:`.Configuration` using the values provided.

        Also optionally store a list of :class:`.Result` for this Configuration using the provided `score`.

        This is useful for giving OPTaaS a "warm start" by providing some examples of good/bad Configurations.

        Optionally, it is possible to pass the input as single unwrapped values.

        Args:
            values (Dict, optional): Values assigned to each :class:`.Parameter`. See :attr:`.Configuration.values`.
            score (ScoreValueOrDict, optional): Score obtained when using this Configuration.
                For multi-objective tasks, a dictionary of scores for each objective, using Objective.id values as keys, e.g.
                `{"id1": 1.23, "id2": 4.56}`.
            variance (VarianceValueOrDict >=0, optional, defaults to 0): Variance associated with the score.
                For multi-objective tasks, a dictionary of variances for each objective, using Objective.id values as keys, e.g.
                `{"id1": 1.23, "id2": 4.56}`.
            user_defined_data (Any, optional, ignored if `score` is not provided): Any other data to store in the Result.
            configurations (List[UserDefinedConfiguration], optional): list of UserDefinedConfiguration.

            Either values or configurations should be provided.

        Returns:
            The newly created list of :class:`.Configuration`.
            In case a single unwrapped value was provided, it returns the unwrapped configuration.

        Raises:
            :class:`.OPTaaSError` if the values are otherwise invalid or the server is unavailable.
        """
        if (values is None) == (configurations is None):
            raise ValueError("Either one of `values` or `configurations` should be provided, but not both.")

        is_unwrapped = configurations is None
        if is_unwrapped:
            configurations = [UserDefinedConfiguration(values, score, variance, user_defined_data)] # type: ignore
        body = {"configurations": [c.get_body() for c in configurations]} # type: ignore

        response = self._session.post(self._configurations_url, body)
        returned_configurations = [Configuration(c) for c in response.body['configurations']]
        if is_unwrapped:
            return returned_configurations[0]
        return returned_configurations

    def record_result(self, configuration: Configuration, score: ScoreValueOrDict = None, error: str = None,
                      variance: VarianceValueOrDict = None, user_defined_data: Any = None) -> Configuration:
        """Make a POST request to OPTaaS to record a :class:`.Result` for the given :class:`.Configuration`.

        Must specify **either** `score` **or** `error`.

        After the Result is recorded, OPTaaS will automatically generate the next Configuration for you to try.

        Args:
            configuration (Configuration): The Configuration used to obtain this Result.
            score (ScoreValueOrDict): The score obtained.
                For multi-objective tasks, a dictionary of scores for each objective, using Objective.id values as keys, e.g.
                `{"id1": 1.23, "id2": 4.56}`.
            error (Any): Any data related to an error encountered while calculating the score.
            variance (VarianceValueOrDict >=0, optional, defaults to 0): Variance associated with the score.
                For multi-objective tasks, a dictionary of variances for each objective, using Objective.id values as keys, e.g.
                `{"id1": 1.23, "id2": 4.56}`.
            user_defined_data (Any, optional): Any other data you wish to store as part of this Result.

        Returns:
            The next :class:`.Configuration` generated by OPTaaS.

        Raises:
            :class:`.OPTaaSError` if the data provided is invalid or the server is unavailable.
            :class:`.ValueError` if both score and error are provided, or neither.
        """
        result = Result(configuration=configuration, score=score, error=error, variance=variance,
                        user_defined_data=user_defined_data)
        response = self._session.post(configuration.results_url, result.to_json_without_configuration())
        return Configuration(response.body['nextConfiguration'])

    def record_results(self, results: List[Result]) -> List[Configuration]:
        """Make a POST request to OPTaaS to store a batch of :class:`Results <.Result>` and get the next batch of :class:`Configurations <.Configuration>`.

        Args:
            results (List[Result]): List of Results to record. Must be non-empty.

        Returns:
            A list of the same length as `results`, containing the next :class:`Configurations <.Configuration>` for you to try.

        Raises:
            :class:`.OPTaaSError` if the data provided is invalid or the server is unavailable.
            :class:`.ValueError` if the result list is empty.
        """
        if not results:
            raise ValueError("Result list must be non-empty.")

        body = {'results': [result.to_json() for result in results]}
        response = self._session.post(self._results_url, body)
        return [Configuration(json) for json in response.body['nextConfigurations']]

    def get_configurations(self, limit: int = None) -> List[Configuration]:
        """Make a GET request to OPTaaS to retrieve a list of :class:`Configurations <.Configuration>` for this Task.

        Args:
            limit (int, optional, minimum 1): Upper bound on the number of Configurations that will be returned.

        Returns:
            The list of :class:`Configurations <.Configuration>`.

        Raises:
            :class:`.OPTaaSError` if the limit is invalid or the server is unavailable.
        """
        url = self._configurations_url
        if limit is not None:
            url += f'?limit={limit}'
        response = self._session.get(url)
        return [Configuration(json) for json in response.body['configurations']]

    def get_configuration(self, configuration_id: str) -> Configuration:
        """Make a GET request to OPTaaS to retrieve a specific :class:`.Configuration` by id.

        Args:
            configuration_id (str): Unique id for the Configuration.

        Returns:
            The retrieved :class:`.Configuration`.

        Raises:
            :class:`.OPTaaSError` if the configuration_id is invalid or the server is unavailable.
        """
        url = f'{self._configurations_url}/{configuration_id}'
        response = self._session.get(url)
        return Configuration(response.body)

    def get_results(self, limit: int = None, best_first: bool = False, as_dataframe: bool = False,
                    include_configurations=None) -> Union[List[StoredResult], 'DataFrame']:
        """Make a GET request to OPTaaS to retrieve a list of :class:`Results <.StoredResult>` for this Task.

        Args:
            limit (int, optional, minimum 1): Upper bound on the number of Results that will be returned.
            best_first (bool, optional, default False):
                If True, Results will appear in score order, with the best score first (not currently supported for multi-objective tasks).
                If False, Results will appear in the order they were created.
            as_dataframe (bool, optional, default False):
                Return the data as a Pandas DataFrame. It will include a column for each parameter, plus the
                score, variance and error from each Result.
            include_configurations: Deprecated.

        Returns:
            The list of :class:`Results <.StoredResult>` or a DataFrame.

        Raises:
            :class:`.OPTaaSError` if the limit is invalid or the server is unavailable.
        """
        if include_configurations is not None:
            raise ValueError("include_configurations has been deprecated. Results will now always include the Configuration.")

        if best_first and self.json.get('objectives'):
            raise ValueError(_MULTI_OBJECTIVE_UNSUPPORTED_ERROR)

        url = self._results_url
        query_params: Dict[str, Any] = {}
        if limit is not None:
            query_params['limit'] = limit
        if best_first:
            query_params['order'] = 'bestFirst'
        response = self._session.get(url, query_params=query_params)
        results = self._make_stored_results(response)
        if as_dataframe:
            from pandas.io.json import json_normalize  # pylint: disable=import-outside-toplevel
            return json_normalize([result.as_pandas_row() for result in results])
        return results

    def _make_stored_results(self, response: OPTaaSResponse) -> List[StoredResult]:  # pylint: disable=no-self-use
        return [StoredResult(json) for json in response.body['results']]

    def get_result(self, result_id: int) -> StoredResult:
        """Make a GET request to OPTaaS to retrieve a specific :class:`.StoredResult` by id.

        Args:
            result_id (str): Unique id for the Result.

        Returns:
            The retrieved :class:`.StoredResult`.

        Raises:
            :class:`.OPTaaSError` if the result_id is invalid or the server is unavailable.
        """
        url = f'{self._results_url}/{result_id}'
        response = self._session.get(url)
        return StoredResult(response.body)

    def get_best_result_and_configuration(self) -> StoredResult:
        """Make a GET request to OPTaaS to retrieve the Result with the best score, including the Configuration used to obtain it.

        Not currently supported for multi-objective Tasks.

        Returns:
            The best :class:`.StoredResult` with included :class:`.Configuration`.

        Raises:
            :class:`.OPTaaSError` if the server is unavailable.
            :class:`.ValueError` if no results have been posted for this task.
        """
        results = self.get_results(best_first=True, limit=1)
        if results:
            return results[0]
        raise ValueError('No results available for this task yet')

    def get_targets_estimates(self):
        """Make a GET request to OPTaaS to retrieve the current estimates for the targets of this task.

        Not supported for tasks without targets.

        Returs:
            The current estimates for the targets of this task as dictionary from target id to estimate."""
        if not self.json.get('targets'):
            raise ValueError("Targets estimates is provided only for tasks with targets.")
        response = self._session.get(self._targets_estimates_url)
        return response.body["results"]

    def get_sqrt_crlb(self):
        """Make a GET request to OPTaaS to retrieve the square roots of the Cramer-Rao lower bounds of the estimates.
        https://en.wikipedia.org/wiki/Cram%C3%A9r%E2%80%93Rao_bound
        The value can be used as a proxy of the quality of the estimate.

        Not supported for tasks without targets.

        Returs:
            The square root of the CRLB for the targets of this task as dictionary from target id to sqrt_crlb."""
        if not self.json.get('targets'):
            raise ValueError("The square roots of the Cramer-Rao lower bounds are provided only for tasks with targets.")
        response = self._session.get(self._sqrt_crlb_url)
        return response.body["results"]

    def get_pareto_set(self) -> List[StoredResult]:
        """Make a GET request to OPTaaS to retrieve the set of Pareto front Results for a multi-objective Task.

        These are the Results where, for each objective, the score cannot be improved without reducing the score for another objective.

        Not supported for single-objective Tasks.

        Note: if a Result doesn't contain a score value for all Objectives, it will be excluded from the Pareto set.

        Returns:
            The list of Pareto front Results.

        Raises:
            :class:`.OPTaaSError` if the server is unavailable or the task is not multi-objective.
        """
        if not self.json.get('objectives'):
            raise ValueError(_SINGLE_OBJECTIVE_UNSUPPORTED_ERROR)
        response = self._session.get(self._pareto_url)
        return self._make_stored_results(response)

    def get_surrogate_predictions(self, configurations: List[Dict]) -> List[Prediction]:
        """Make a POST request to OPTaaS to retrieve the surrogate prediction for some configurations

        Args:
            configurations (List[Dict]): key value pairs corresponding to parameter names and their values.
            See :attr:`.Configuration.values`.

        Returns:
            :List[Prediction]: a list of objects, each containing the mean and variance of the surrogate at each
            corresponding configuration point.
            For multi-objective Tasks, the mean and variance are dictionaries containing mean (and
            variance respectively) for each objective.

        Raises:
            :class:`.OPTaaSError` if the configurations are invalid or the server is unavailable
        """
        body = {'configurations': [{'values': configuration} for configuration in configurations]}
        response = self._session.post(self._predictions_url, body)
        predictions = response.body['predictions']
        return [Prediction(json=prediction) for prediction in predictions]

    def complete(self):
        """Make a PUT request to OPTaaS to complete the task (no further configurations or results can be created)"""
        self._session.put(self._complete_url, {})

    def resume(self):
        """Make a PUT request to OPTaaS to resume a completed task"""
        self._session.put(self._resume_url, {})

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __repr__(self):
        return _pprint(self, 'id', 'title', 'user_defined_data', 'parameters', 'constraints')



from enum import Enum


class Goal(Enum):
    """Specifies whether OPTaaS will aim for the lowest (min) or highest (max) score"""
    min = 0  # pylint: disable=invalid-name
    max = 1  # pylint: disable=invalid-name



import abc


class Expression(abc.ABC):
    """Allows us to build expressions for :class:`Constraints <.Constraint>` using common operators.
    The following operators are supported:
        * addition: `x + y`
        * subtraction: `x - y`
        * multiplication: `x * y`
        * division: `x / y`
        * floor division: `x // y`
        * remainder: `x % y`
        * exponentiation: `x ** y`
        * equality: `x == y`
        * inequality: `x != y`
        * greater than: `x > y`
        * greater than or equal: `x >= y`
        * less than: `x < y`
        * less than or equal: `x <= y`
        * and: `x & y`
        * or: `x | y`

    Example:
        `((param1 + param2 <= param3) & param4.is_present()) | param5 == 'abc'`

    Note the use of brackets with the `&` and `|` operators - this is because they don't have the same precedence as
    `and` and `or` so you need to be explicit in order to generate the correct expression.


    See `this page <../constraints.ipynb>`_ for more examples.

    Refer to the OPTaaS swagger page for more details on constraint syntax.
    """

    def __add__(self, other) -> 'Expression':
        return BinaryPredicate(self, '+', other)

    def __radd__(self, other) -> 'Expression':
        return BinaryPredicate(other, '+', self)

    def __sub__(self, other) -> 'Expression':
        return BinaryPredicate(self, '-', other)

    def __rsub__(self, other) -> 'Expression':
        return BinaryPredicate(other, '-', self)

    def __mul__(self, other) -> 'Expression':
        return BinaryPredicate(self, '*', other)

    def __rmul__(self, other) -> 'Expression':
        return BinaryPredicate(other, '*', self)

    def __truediv__(self, other) -> 'Expression':
        return BinaryPredicate(self, '/', other)

    def __rtruediv__(self, other) -> 'Expression':
        return BinaryPredicate(other, '/', self)

    def __floordiv__(self, other) -> 'Expression':
        return BinaryPredicate(self, '//', other)

    def __rfloordiv__(self, other) -> 'Expression':
        return BinaryPredicate(other, '//', self)

    def __mod__(self, other) -> 'Expression':
        return BinaryPredicate(self, '%', other)

    def __rmod__(self, other) -> 'Expression':
        return BinaryPredicate(other, '%', self)

    def __pow__(self, other) -> 'Expression':
        return BinaryPredicate(self, '**', other)

    def __rpow__(self, other) -> 'Expression':
        return BinaryPredicate(other, '**', self)

    def __eq__(self, other) -> 'Expression':  # type: ignore
        return BinaryPredicate(self, '==', other)

    def __ne__(self, other) -> 'Expression':  # type: ignore
        return BinaryPredicate(self, '!=', other)

    def __gt__(self, other) -> 'Expression':
        return BinaryPredicate(self, '>', other)

    def __lt__(self, other) -> 'Expression':
        return BinaryPredicate(self, '<', other)

    def __ge__(self, other) -> 'Expression':
        return BinaryPredicate(self, '>=', other)

    def __le__(self, other) -> 'Expression':
        return BinaryPredicate(self, '<=', other)

    def __and__(self, other) -> 'Expression':
        return BinaryPredicate(self, '&&', other)

    def __or__(self, other) -> 'Expression':
        return BinaryPredicate(self, '||', other)

    @abc.abstractmethod
    def to_optaas_expression(self) -> str:
        pass


class Predicate(Expression):
    """An Expression that evaluates to a boolean value that can be used in a :class:`.Constraint`"""

    def __init__(self, optaas_expression: str) -> None:
        self._optaas_expression = optaas_expression

    def to_optaas_expression(self) -> str:
        """A string representation of this expression in a format which can be parsed by OPTaaS"""
        return self._optaas_expression


class BinaryPredicate(Predicate):
    """A Predicate with 2 operands"""

    def __init__(self, left_operand, operator: str, right_operand) -> None:
        super().__init__(f'{_format(left_operand)} {operator} {_format(right_operand)}')


class UnaryPredicate(Predicate):
    """A Predicate with a single operand"""

    def __init__(self, operand, operator: str) -> None:
        super().__init__(f'{_format(operand)} {operator}')


class WhenThenExpression(abc.ABC):
    """
    Defines a when expression, then expression clause:

    The "when" part is optional (if omitted, the then part applies).

    NOTE: At the MFOpt level, WhenThenExpressions get transformed to either constraints or prior mean functions, which
    means the "then" expressions have to conform to some formats: they have to be boolean for constraints, and not
    boolean for prior means.
    """
    def __init__(self, then: Expression, when: Expression = None) -> None:
        self.then = then
        self.when = when

    def to_optaas_expression(self) -> str:
        return (self.then.to_optaas_expression() if self.when is None
                else f'if {self.when.to_optaas_expression()} then {self.then.to_optaas_expression()}')

    def __eq__(self, other):
        return isinstance(other, WhenThenExpression) and self.to_optaas_expression() == other.to_optaas_expression()

    def __repr__(self):
        return self.to_optaas_expression()


class Constraint(WhenThenExpression):
    """
    This is an alias for the WhenThenExpression class which defines a when expression, then expression clause:

    The "when" part is optional (if omitted, the then part applies).
    """
    def __eq__(self, other):
        return isinstance(other, Constraint) and self.to_optaas_expression() == other.to_optaas_expression()


class PriorMeanExpression(WhenThenExpression):
    """
    This is an alias for the WhenThenExpression class which defines a when expression, then expression clause:

    The "when" part is optional (if omitted, the then part applies).
    """
    def __eq__(self, other):
        return isinstance(other, PriorMeanExpression) and self.to_optaas_expression() == other.to_optaas_expression()


def _format(operand):
    from mindfoundry.optaas.client.parameter import Parameter  # pylint: disable=import-outside-toplevel
    if operand is True:
        operand = 'true'
    elif operand is False:
        operand = 'false'
    elif isinstance(operand, Parameter):
        operand = operand.to_optaas_expression()
    elif isinstance(operand, str):
        operand = f"'{operand}'"
    elif isinstance(operand, BinaryPredicate):
        operand = '( ' + operand.to_optaas_expression() + ' )'
    elif isinstance(operand, UnaryPredicate):
        operand = operand.to_optaas_expression()
    return operand



from typing import Dict, Any


class Prediction:
    """The result obtained by taking a :class:`.Configuration` and running it through the surrogate model.

    Must contain a `mean` value.
    You cannot specify `variance` without `mean`.

    Attributes:
        json (Dict): The full Prediction object as stored in OPTaaS in JSON format.
        configuration (Dict[str, Any]): key-value pairs corresponding to the parameters and their respective values
        mean (float): Mean from surrogate model prediction
        variance (float): Variance of the surrogate model prediction. Must be >= 0.
    """

    def __init__(self, json: Dict[str, Any]) -> None:
        self.mean = json['mean']
        self.variance = json['variance']
        self.json = json
        self.configuration = json['configuration']



from typing import Dict

from requests import Session, Response
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from mindfoundry.optaas.client.version_check import compare_versions, CLIENT_VERSION

DEFAULT_MAX_RETRIES = 3
RETRY_STATUS_CODES = [500, 501, 502, 503, 504]
DEFAULT_TIMEOUT = 60


class OPTaaSError(RuntimeError):
    """Wrapper class for an error :class:`.Response` received from OPTaaS."""

    def __init__(self, response: Response) -> None:
        self.status_code = response.status_code
        try:
            self.message = response.json()['message']
        except BaseException:
            self.message = response.content.decode() if response.content else ''
        super().__init__(f"Status: {self.status_code}  Message: {self.message}")


class OPTaaSResponse:
    """Wrapper class for a successful :class:`.Response` received from OPTaaS."""

    def __init__(self, response: Response, disable_version_check: bool = False) -> None:
        if not disable_version_check:
            compare_versions(response)
        if response.ok:
            self.body = response.json()
        else:
            raise OPTaaSError(response)


class OPTaaSSession:
    """Wrapper class for a :class:`.Session` that makes requests to OPTaaS.

    Args:
        server_url (str): URL of your OPTaaS server
        api_key (str): Your personal API key
        disable_version_check (bool, default False):
            Set to True if you don't want to be notified when a new version of the client is available
        max_retries (int, default 3): How many times to retry a request if a connection error occurs.
        keep_alive (bool, default True): Whether to set the `Connection` HTTP Header to "keep-alive".
    """

    def __init__(self, server_url: str, api_key: str, disable_version_check: bool, max_retries: int, keep_alive: bool) -> None:
        self._session = self._make_session(server_url, max_retries)
        self._root_url = server_url
        self._headers = {
            'X-ApiKey': api_key,
            'User-Agent': 'PythonClient/' + CLIENT_VERSION,
            'Connection': 'keep-alive' if keep_alive else 'close'
        }
        self._disable_version_check = disable_version_check

    @staticmethod
    def _make_session(server_url: str, max_retries: int) -> Session:
        session = Session()
        retry = Retry(total=max_retries, connect=max_retries, backoff_factor=0.5, method_whitelist=None,
                      status_forcelist=RETRY_STATUS_CODES)
        session.mount(server_url, HTTPAdapter(max_retries=retry))
        return session

    def post(self, endpoint: str, body: dict, timeout: float = DEFAULT_TIMEOUT) -> OPTaaSResponse:
        """Make a POST request to OPTaaS with a JSON body.

        Args:
            endpoint (str): Endpoint for the request (will be appended to the server_url).
            body (dict): Request body in JSON format.
            timeout (float, optional): Request timeout in seconds

        Returns:
            The :class:`.OPTaaSResponse` to the request.

        Raises:
            :class:`.OPTaaSError` if an error response is received.
        """
        response = self._session.post(self._root_url + endpoint, json=body, headers=self._headers, timeout=timeout)
        return OPTaaSResponse(response, disable_version_check=self._disable_version_check)

    def put(self, endpoint: str, body: dict, timeout: float = DEFAULT_TIMEOUT) -> OPTaaSResponse:
        """Make a PUT request to OPTaaS with a JSON body.

        Args:
            endpoint (str): Endpoint for the request (will be appended to the server_url).
            body (dict): Request body in JSON format.
            timeout (float, optional): Request timeout in seconds

        Returns:
            The :class:`.OPTaaSResponse` to the request.

        Raises:
            :class:`.OPTaaSError` if an error response is received.
        """
        response = self._session.put(self._root_url + endpoint, json=body, headers=self._headers, timeout=timeout)
        return OPTaaSResponse(response, disable_version_check=self._disable_version_check)

    def get(self, endpoint: str, query_params: Dict = None, timeout: float = DEFAULT_TIMEOUT) -> OPTaaSResponse:
        """Make a GET request to OPTaaS

        Args:
            endpoint (str): Endpoint for the request (will be appended to the server_url).
            query_params (Dict, optional): Query parameters in Dict format (will be appended to the url).
            timeout (float, optional): Request timeout in seconds

        Returns:
            The :class:`.OPTaaSResponse` to the request.

        Raises:
            :class:`.OPTaaSError` if an error response is received.
        """
        response = self._session.get(self._root_url + endpoint, headers=self._headers, params=query_params, timeout=timeout)
        return OPTaaSResponse(response, disable_version_check=self._disable_version_check)

    def delete(self, endpoint: str, timeout: float = DEFAULT_TIMEOUT) -> OPTaaSResponse:
        """Make a DELETE request to OPTaaS

        Args:
            endpoint (str): Endpoint for the request (will be appended to the server_url).
            timeout (float, optional): Request timeout in seconds

        Returns:
            The :class:`.OPTaaSResponse` to the request.

        Raises:
            :class:`.OPTaaSError` if an error response is received.
        """
        response = self._session.delete(self._root_url + endpoint, headers=self._headers, timeout=timeout)
        return OPTaaSResponse(response, disable_version_check=self._disable_version_check)

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self._root_url == other._root_url  # pylint: disable=protected-access



from dataclasses import dataclass
from typing import Any, Dict, Optional

from mindfoundry.optaas.client.result import ScoreValueOrDict, VarianceValueOrDict, Result
from mindfoundry.optaas.client.configuration import Configuration

class _MockConfiguration(Configuration):
    """Used only internally, to post a result for a user-defined configuration."""

    def __init__(self):  # pylint: disable=super-init-not-called
        pass

@dataclass
class UserDefinedConfiguration:
    """ A user defined configuration containing the parameters values and optionally score,
        variance and user_defined_data."""
    values: Dict
    score: Optional[ScoreValueOrDict] = None
    variance: Optional[VarianceValueOrDict] = None
    user_defined_data: Any = None

    def get_body(self) -> Dict:
        body = {"values": self.values}
        if self.score:
            result = Result(configuration=_MockConfiguration(), score=self.score, variance=self.variance,
                            user_defined_data=self.user_defined_data)
            body['results'] = [result.to_json_without_configuration()] # type: ignore
        return body



from typing import Sequence

from sklearn.ensemble import \
    RandomForestClassifier as BaseRandomForestClassifier, \
    ExtraTreesClassifier as BaseExtraTreesClassifier, \
    RandomForestRegressor as BaseRandomForestRegressor, \
    ExtraTreesRegressor as BaseExtraTreesRegressor, \
    GradientBoostingClassifier as BaseGradientBoostingClassifier, \
    GradientBoostingRegressor as BaseGradientBoostingRegressor
from sklearn.tree import DecisionTreeClassifier as BaseDecisionTreeClassifier, \
    DecisionTreeRegressor as BaseDecisionTreeRegressor

from mindfoundry.optaas.client.parameter import Distribution
from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersConstraintsAndPriorMeans, SMALLEST_NUMBER_ABOVE_ZERO


class _OptimizableEnsemble(OptimizableBaseEstimator):
    """Superclass for optimizable ensemble estimators."""

    # Properties must be specified in subclasses
    criterion_values: Sequence[str] = None  # type: ignore
    min_samples_split_distribution: Distribution = None  # type: ignore
    max_depth_distribution: Distribution = None  # type: ignore
    min_impurity_decrease_maximum: float = None  # type: ignore

    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs)\
            -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class: `PriorMeans <.PriorMeans>` to optimize ensemble estimators."""

        parameters = [
            sk.CategoricalParameter('max_features', values=['auto', 'sqrt', 'log2']),
            sk.IntParameter('min_samples_split', minimum=2, maximum=20,
                            distribution=self.min_samples_split_distribution),
            sk.IntParameter('min_samples_leaf', minimum=1, maximum=20),
            sk.CategoricalParameter('criterion', values=self.criterion_values),
            sk.IntParameter('max_leaf_nodes', minimum=10, maximum=10000, optional=True,
                            distribution=Distribution.LOGUNIFORM),
            sk.IntParameter('max_depth', minimum=1, maximum=100, optional=True,
                            distribution=self.max_depth_distribution),  # type: ignore
            sk.FloatParameter('min_weight_fraction_leaf', minimum=0.0, maximum=0.5),
            sk.FloatParameter('min_impurity_decrease', minimum=0, maximum=self.min_impurity_decrease_maximum),  # type: ignore
        ]
        return parameters, [], []


class _OptimizableForestOrTree(_OptimizableEnsemble):
    """Superclass for optimizable forest/tree-based estimators."""

    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
          and :class:`PriorMeans <.PriorMeans>` to optimize forest/tree-based estimators."""

        parameters, constraints, prior_means = super().make_parameters_constraints_and_prior_means(sk, **kwargs)


        parameters.extend([
            sk.BoolParameter('bootstrap'),
            sk.IntParameter('n_estimators', minimum=10, maximum=500),
        ])

        return parameters, constraints, prior_means


class _OptimizableEnsembleClassifier(_OptimizableEnsemble):
    """Superclass for optimizable ensemble classifiers."""

    criterion_values = ['gini', 'entropy']
    min_impurity_decrease_maximum = 1


class _OptimizableEnsembleRegressor(_OptimizableEnsemble):
    """Superclass for optimizable ensemble regressors."""

    criterion_values = ['absolute_error', 'squared_error', 'poisson', 'mse', 'mae']
    min_impurity_decrease_maximum = 0.01


class _OptimizableRandomForest(_OptimizableForestOrTree):
    """Superclass for optimizable RandomForest estimators."""

    min_samples_split_distribution = Distribution.UNIFORM
    max_depth_distribution = Distribution.LOGUNIFORM


class _OptimizableExtraTrees(_OptimizableForestOrTree):
    """Superclass for optimizable ExtraTrees estimators."""

    min_samples_split_distribution = Distribution.LOGUNIFORM
    max_depth_distribution = Distribution.UNIFORM


class RandomForestClassifier(BaseRandomForestClassifier, _OptimizableEnsembleClassifier, _OptimizableRandomForest):
    """Allows us to optimize :class:`.RandomForestClassifier` estimators."""


class ExtraTreesClassifier(BaseExtraTreesClassifier, _OptimizableEnsembleClassifier, _OptimizableExtraTrees):
    """Allows us to optimize :class:`.ExtraTreesClassifier` estimators."""


class RandomForestRegressor(BaseRandomForestRegressor, _OptimizableEnsembleRegressor, _OptimizableRandomForest):
    """Allows us to optimize :class:`.RandomForestRegressor` estimators."""


class ExtraTreesRegressor(BaseExtraTreesRegressor, _OptimizableEnsembleRegressor, _OptimizableExtraTrees):
    """Allows us to optimize :class:`.ExtraTreesRegressor` estimators."""


class _OptimizableDecisionTree(_OptimizableEnsemble):
    """Superclass for optimizable DecisionTree estimators."""

    min_samples_split_distribution = Distribution.LOGUNIFORM
    max_depth_distribution = Distribution.UNIFORM

    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class:`PriorMeans <.PriorMeans>` to optimize DecisionTree-based estimators."""

        parameters, constraints, prior_means = super().make_parameters_constraints_and_prior_means(sk, **kwargs)

        parameters.extend([
            sk.CategoricalParameter('splitter', values=["best", "random"]),
        ])

        return parameters, constraints, prior_means


class DecisionTreeClassifier(BaseDecisionTreeClassifier, _OptimizableDecisionTree, _OptimizableEnsembleClassifier):
    """Allows us to optimize a :class:`.DecisionTreeClassifier`."""


class DecisionTreeRegressor(BaseDecisionTreeRegressor, _OptimizableDecisionTree):
    """Allows us to optimize a :class:`.DecisionTreeRegressor`."""

    criterion_values = ["mse", "friedman_mse", "mae", "squared_error"]
    min_impurity_decrease_maximum = 0.01


class _OptimizableGradientBoosting(_OptimizableEnsemble):
    """Superclass for optimizable GradientBoosting estimators."""

    criterion_values = ["mse", "friedman_mse", "mae", "squared_error"]
    min_impurity_decrease_maximum = 1
    min_samples_split_distribution = Distribution.UNIFORM
    max_depth_distribution = Distribution.UNIFORM

    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class:`PriorMeans <.PriorMeans>` to optimize a GradientBoosting estimator."""

        parameters, constraints, prior_means = super().make_parameters_constraints_and_prior_means(sk, **kwargs)

        parameters.extend([
            sk.FloatParameter('learning_rate', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1),
            sk.IntParameter('n_estimators', minimum=10, maximum=500),
            sk.FloatParameter('subsample', minimum=0, maximum=1),
        ])

        return parameters, constraints, prior_means


class GradientBoostingClassifier(BaseGradientBoostingClassifier, _OptimizableGradientBoosting):
    """Allows us to optimize a :class:`.GradientBoostingClassifier`."""

    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class:`PriorMeans <.PriorMeans>` and :class:`Constraints <.Constraint>` to optimize
         a :class:`.GradientBoostingClassifier`.

        Args:
            class_count (int, optional): Number of classes in the classification dataset.
                If set to a number other than `2`, the `loss` parameter will not be optimized (because it can only be set to "deviance").
        """

        parameters, constraints, prior_means = super().make_parameters_constraints_and_prior_means(sk, **kwargs)

        if kwargs.get('class_count') == 2:
            parameters.append(sk.CategoricalParameter('loss', values=["deviance", "exponential"]))

        return parameters, constraints, prior_means

    def _make_estimator(self, append=True):
        pass  # BaseGradientBoostingClassifier doesn't implement this either


class GradientBoostingRegressor(BaseGradientBoostingRegressor, _OptimizableGradientBoosting):
    """Allows us to optimize a :class:`.GradientBoostingRegressor`."""

    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class:`PriorMeans <.PriorMeans>` to optimize a :class:`.GradientBoostingClassifier`."""

        parameters, constraints, prior_means = super().make_parameters_constraints_and_prior_means(sk, **kwargs)

        parameters.extend([
            sk.CategoricalParameter('loss', values=["squared_error", "absolute_error", "huber", "quantile"]),
            sk.FloatParameter('alpha', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1),
        ])

        return parameters, constraints, prior_means

    def _make_estimator(self, append=True):
        pass  # BaseGradientBoostingClassifier doesn't implement this either



from sklearn.linear_model import Lasso as BaseLasso, Ridge as BaseRidge

from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersConstraintsAndPriorMeans, SMALLEST_NUMBER_ABOVE_ZERO


class _OptimizableLinearModel(OptimizableBaseEstimator):
    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs)\
            -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class:`PriorMeans <.PriorMeans>` to optimize a linear model."""

        return [sk.FloatParameter('alpha', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1)], [], []


class Lasso(BaseLasso, _OptimizableLinearModel):
    """Allows us to optimize a :class:`.Lasso` estimator."""


class Ridge(BaseRidge, _OptimizableLinearModel):
    """Allows us to optimize a :class:`.Ridge` estimator."""



from sklearn.decomposition import PCA as BasePCA

from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersConstraintsAndPriorMeans


class PCA(BasePCA, OptimizableBaseEstimator):
    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs)\
            -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class:`PriorMeans <.PriorMeans>` to optimize a :class:`.PCA` estimator.

        Args:
            feature_count (int): Total number of features in your dataset.
        """

        feature_count = self.get_required_kwarg(kwargs, 'feature_count')
        max_n_components = feature_count - 1 if self.svd_solver == 'arpack' else feature_count

        return [
            sk.IntParameter('n_components', minimum=1, maximum=max_n_components),
            sk.BoolParameter('whiten')
        ], [], []



from enum import Enum
from typing import Dict

from mindfoundry.optaas.client.session import OPTaaSSession
from mindfoundry.optaas.client.utils import _pprint


class UserRole(Enum):
    STANDARD = "standard"  # Can use all non-admin endpoints except API key-related
    ADMIN = "admin"  # Can use all endpoints including admin ones (API key-related)
    READ_ONLY = "read_only"  # Can use only non-admin GET endpoints


class ApiKey:
    """A key that can be used to interact with OPTaaS.

    Attributes:
        json (Dict): The full JSON representation of this key in OPTaaS.
        id (str): The actual API key.
        role (UserRole): The role assigned to this key.
        expired (bool): Whether this key has expired, i.e. can no longer be used.
    """

    def __init__(self, json: Dict, session: OPTaaSSession) -> None:
        self.json = json
        self.id = json['id']  # pylint: disable=invalid-name
        self.role = json['role']
        self.expired = json['expired']
        self._url = json['_links']['self']['href']
        self._session = session

    def set_role(self, role: UserRole) -> None:
        """Modifies the role associated with an API Key by making a PUT request to OPTaaS. Only available to Admin users."""
        self._session.put(self._url, {"role": role.value})
        self.role = role.value
        self.json['role'] = role.value

    def set_expired(self, expired: bool) -> None:
        """Expires or un-expires an API Key by making a PUT request to OPTaaS. Only available to Admin users."""
        self._session.put(self._url, {"expired": expired})
        self.expired = expired
        self.json['expired'] = expired

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.__dict__ == other.__dict__

    def __repr__(self):
        return _pprint(self, 'id', 'role', 'expired')



from xgboost.sklearn import XGBClassifier as BaseXGBClassifier

from mindfoundry.optaas.client.sklearn_pipelines.mixin import OptimizableBaseEstimator
from mindfoundry.optaas.client.sklearn_pipelines.parameter_maker import SklearnParameterMaker
from mindfoundry.optaas.client.sklearn_pipelines.utils import ParametersConstraintsAndPriorMeans, SMALLEST_NUMBER_ABOVE_ZERO


class XGBClassifier(BaseXGBClassifier, OptimizableBaseEstimator):
    def make_parameters_constraints_and_prior_means(self, sk: SklearnParameterMaker, **kwargs) -> ParametersConstraintsAndPriorMeans:
        """Generates :class:`Parameters <.Parameter>`, :class:`Constraints <.Constraint>`
         and :class:`PriorMeans <.PriorMeans>` to optimize a :class:`.XGBClassifier` estimator.

        Args:
            gpu_enabled (bool, optional, default=False):
                If True, the `objective` parameter will include gpu-specific values such as 'gpu:reg:linear'.
        """

        objective_values = [
            'reg:linear', 'reg:logistic', 'binary:logistic', 'binary:logitraw', 'count:poisson',
            'survival:cox', 'multi:softmax', 'multi:softprob', 'rank:pairwise', 'reg:gamma', 'reg:tweedie'
        ]

        if kwargs.get('gpu_enabled'):
            objective_values += ['gpu:reg:linear', 'gpu:reg:logistic', 'gpu:binary:logistic', 'gpu:binary:logitraw']

        parameters = [
            sk.CategoricalParameter('objective', values=objective_values),
            sk.CategoricalParameter('booster', values=["gbtree", "gblinear", "dart"]),
            sk.FloatParameter('learning_rate', minimum=0, maximum=1),
            sk.FloatParameter('gamma', minimum=0, maximum=20),
            sk.IntParameter('max_depth', minimum=0, maximum=20),
            sk.IntParameter('min_child_weight', minimum=0, maximum=10),
            sk.IntParameter('max_delta_step', minimum=0, maximum=10),
            sk.FloatParameter('subsample', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1),
            sk.FloatParameter('colsample_bytree', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1),
            sk.FloatParameter('colsample_bylevel', minimum=SMALLEST_NUMBER_ABOVE_ZERO, maximum=1),
            sk.FloatParameter('reg_lambda', minimum=0, maximum=1),
            sk.FloatParameter('reg_alpha', minimum=0, maximum=1),
            sk.FloatParameter('scale_pos_weight', minimum=0, maximum=1),
            sk.FloatParameter('base_score', minimum=0, maximum=1),
            sk.IntParameter('n_estimators', minimum=5, maximum=1000),
        ]
        return parameters, [], []



import copy
from typing import List, Dict, Callable, Union, TYPE_CHECKING

from sklearn import clone
from sklearn.pipeline import Pipeline

from mindfoundry.optaas.client.configuration import Configuration
from mindfoundry.optaas.client.result import StoredResult, ScoringFunctionOutput, ScoreValueOrDict
from mindfoundry.optaas.client.session import OPTaaSResponse
from mindfoundry.optaas.client.sklearn_pipelines.mixin import EstimatorChoice, OptimizablePipeline
from mindfoundry.optaas.client.sklearn_pipelines.utils import EstimatorTuple, Estimator
from mindfoundry.optaas.client.task import Task, LoggingLevel
from mindfoundry.optaas.client.utils import move_dict_value_up_one_level, get_first_key

if TYPE_CHECKING:  # pragma: no cover
    from pandas import DataFrame  # pylint: disable=unused-import


class SklearnTask(Task):
    """A Task that can convert a :class:`.Configuration` into a sklearn :class:`.Pipeline`"""

    def __init__(self, task: Task, estimators: List[EstimatorTuple]) -> None:
        self._estimators = estimators
        super().__init__(task.json, task._session)  # pylint: disable=protected-access

    def run(self, scoring_function: Callable[[Pipeline], ScoringFunctionOutput], max_iterations: int,
            score_threshold: ScoreValueOrDict = None, logging_level: LoggingLevel = "INFO") -> Union[StoredResult, List[StoredResult]]:
        """Run this task, using the provided scoring function to calculate the score for each configuration.

        Args:
            scoring_function (Callable[[Pipeline], ScoringFunctionOutput]):
                Function that takes a sklearn Pipeline as input and returns a score or a tuple of (score, variance).
            max_iterations (int):
                Max number of iterations to run, i.e. number of results to record before stopping.
            score_threshold (float, optional):
                Stop running the task when the score is equal to or better than this value.
            logging_level (Literal["DEBUG", "INFO", "WARNING", "ERROR"], optional, defaults to "INFO"): Set the logging level.

        Returns:
            The best recorded :class:`.Result` with the :class:`.Pipeline` that was used to achieve it.
            For multi-objective tasks, the set of Pareto front Results will be returned instead.

        Raises:
            :class:`.OPTaaSError` if the server is unavailable.
        """

        def calculate_score(**configuration_values):
            pipeline = self._make_pipeline(configuration_values['pipeline'])
            return scoring_function(pipeline)

        return super().run(scoring_function=calculate_score, max_iterations=max_iterations,
                           score_threshold=score_threshold, logging_level=logging_level)

    def _make_stored_results(self, response: OPTaaSResponse) -> List[StoredResult]:
        results = super()._make_stored_results(response)
        for result in results:
            if isinstance(result.configuration, Configuration):
                result.pipeline = self.make_pipeline(result.configuration)
        return results

    def make_pipeline(self, configuration: Configuration) -> Pipeline:
        """Creates a sklearn :class:`.Pipeline` and sets its parameters based on the provided :class:`.Configuration`"""
        return self._make_pipeline(configuration.values['pipeline'])

    def _make_pipeline(self, pipeline_values: Dict) -> Pipeline:
        configuration_values = copy.deepcopy(pipeline_values)
        pipeline_parameters = self.parameters[0]['items']
        return self._make_pipeline_from_values(configuration_values, pipeline_parameters, self._estimators)

    def _make_pipeline_from_values(self, values: Dict, parameters: List[Dict], estimators: List[EstimatorTuple]):
        new_estimators = []

        for name, estimator in estimators:
            if name in values:
                estimator_parameter = next(p for p in parameters if p['name'] == name)
                new_estimator = self._make_estimator(name, estimator, values, estimator_parameter)
                new_estimators.append((name, new_estimator))

        return Pipeline(new_estimators)

    def _make_estimator(self, name: str, estimator: Estimator, values: Dict, estimator_parameter: Dict):
        if isinstance(estimator, OptimizablePipeline):
            return self._make_pipeline_from_values(values[name], estimator_parameter['items'], estimator.estimators)

        if isinstance(estimator, EstimatorChoice):
            choice = get_first_key(values[name])
            choice_index = int(choice)
            move_dict_value_up_one_level(values, name)
            estimator = estimator.estimators[choice_index]
            estimator_parameter = estimator_parameter['choices'][choice_index]
            return self._make_estimator(name, estimator, values, estimator_parameter)

        _flatten_values(values[name], estimator_parameter)
        cloned_estimator = clone(estimator)
        cloned_estimator.set_params(**values[name])
        return cloned_estimator


def _flatten_values(values: Dict, parameter_group: Dict) -> None:
    for parameter in parameter_group['items']:
        name = parameter['name']
        if name in values:
            if parameter['type'] == 'choice':
                move_dict_value_up_one_level(values, name)
            elif parameter['type'] == 'group':
                value: Dict = values.get(name)  # type: ignore
                group_values = list(value.values())
                if len(group_values) == 1 and isinstance(group_values[0], Dict):
                    values[name] = group_values[0]
                else:
                    values[name] = group_values
        else:
            values[name] = None
